/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TabLang extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is Andrey Lokot.
 * Portions created by the Initial Developer are Copyright (C) 2010-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrey Lokot <andreylokot@yandex.ru>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var EXPORTED_SYMBOLS = ["TablangKbSwitcherWin32"];

Components.utils.import("resource://gre/modules/ctypes.jsm");

const HKL = ctypes.uintptr_t;

function TablangKbSwitcherWin32() {
	let abi = (Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULRuntime)
				.XPCOMABI.indexOf("x86_64") == -1) ? ctypes.winapi_abi : ctypes.default_abi;
	this.libUser32 = ctypes.open("user32.dll");
	this.libKernel32 = ctypes.open("kernel32.dll");
	this.winapi = {
		GetKeyboardLayout: this.libUser32.declare("GetKeyboardLayout", abi,
			HKL, ctypes.uint32_t),
		ActivateKeyboardLayout: this.libUser32.declare("ActivateKeyboardLayout", abi,
			HKL, HKL, ctypes.uint32_t),
		GetKeyboardLayoutList: this.libUser32.declare("GetKeyboardLayoutList", abi,
			ctypes.uint32_t, ctypes.int32_t, HKL.ptr),
		GetLocaleInfoW: this.libKernel32.declare("GetLocaleInfoW", abi,
			ctypes.int32_t, ctypes.uint32_t, ctypes.uint32_t, ctypes.jschar.ptr, ctypes.int32_t)
	};
}

TablangKbSwitcherWin32.prototype = {
	getCurrentLang: function TablangKbSwitcherWin32_getCurrentLang() {
		return "0x" + this.winapi.GetKeyboardLayout(0).toString(16);
	},
	setCurrentLang: function TablangKbSwitcherWin32_setCurrentLang(newLang) {
		let result = this.winapi.ActivateKeyboardLayout(ctypes.uintptr_t(ctypes.UInt64(newLang)), 0);
		if (result == 0) {
			throw new Error("winapi.ActivateKeyboardLayout failed with newLang = " + newLang);
		}
	},
	getLangList: function TablangKbSwitcherWin32_getLangList() {
		let listLength = this.winapi.GetKeyboardLayoutList(0, HKL.ptr(0));
		if (listLength == 0) {
			throw new Error("winapi.GetKeyboardLayoutList (get length) failed");
		}
		let cList = HKL.array(listLength)();
		if (this.winapi.GetKeyboardLayoutList(listLength, ctypes.cast(cList.address(), HKL.ptr)) == 0) {
			throw new Error("winapi.GetKeyboardLayoutList (get list) failed");
		}
		let jsList = [];
		for (let i = 0; i < listLength; ++i) {
			jsList[i] = "0x" + cList[i].toString(16);
		}
		return jsList;
	},
	getLangName: function TablangKbSwitcherWin32_getLangName(lang) {
		const LOCALE_SLANGUAGE = 0x00000002;
		// lcid = MAKELCID((WORD)lang, SORT_DEFAULT)
		let lcid = ctypes.UInt64.lo(ctypes.UInt64(lang)) & 0xFFFF;
		let bufferLength = this.winapi.GetLocaleInfoW(lcid, LOCALE_SLANGUAGE,
														ctypes.jschar.ptr(0), 0);
		if (bufferLength == 0) {
			throw new Error("winapi.GetLocaleInfoW (get length) failed");
		}
		let buffer = ctypes.jschar.array(bufferLength)();
		if (this.winapi.GetLocaleInfoW(lcid, LOCALE_SLANGUAGE,
				ctypes.cast(buffer.address(), ctypes.jschar.ptr), bufferLength) == 0) {
			throw new Error("winapi.GetLocaleInfoW (get string) failed");
		}
		return buffer.readString();
	}
};

